import csnd

# Class to represent transforming a note 
# by modifying an implicit initial note,
# creating a duration, adding or subtracting pitch,
# adding or subtracting loudness.

class Transform(object):
        def __init__(self, duration, deltaKey, deltaVelocity):
                self.duration = duration
                self.deltaKey = deltaKey
                self.deltaVelocity = deltaVelocity
                self.normalizedDuration = 1.0
                
# Class for generating a piece using a long initial note
# and a set of transforms, recursively layering atop generated notes.

class Generator(object):
        def __init__(self):
                # Create an instance of CppSound for rendering.
                self.csound = csnd.CppSound()
                self.csound.setPythonMessageCallback()
                self.csound.load('../examples/CsoundVST.csd')
                # To contain a list of transforms.
                self.transforms = []
                # Assign instruments to levels (level:instrument)
                self.arrangement = {0:12, 1:4, 2:21, 3:7, 4:37}
                # Assign gains to levels (level:gain)
                self.gains = {0:1.5, 1:1.25, 2:1, 3:1, 4:1}
                # Assign pans to levels (level:pan)
                self.pans = {0:0, 1:-.75, 2:.75, 3:-.5, 4:.5, 5:-.25, 6:.25}
        def addTransform(self, deltaTime, deltaKey, deltaVelocity):
                self.transforms.append(Transform(deltaTime, deltaKey, deltaVelocity))
                self.normalize()
        def normalize(self):
                sum = 0.0
                for transform in self.transforms:
                        sum = sum + transform.duration
                for transform in self.transforms:
                        transform.normalizedDuration = transform.duration / sum
        # Generate a score in the form of a Koch curve.
        # Each note in a generating motive
        # will get a smaller copy of the motive nested atop it,
        # and so on.
        def generate(self, level, levels, initialTime, totalDuration, initialKey, initialVelocity):
                # If the bottom level has already been reached,
                # return without further recursion.
                if level >= levels:
                        return
                time = initialTime
                key = initialKey
                velocity = initialVelocity
                for transform in self.transforms:
                        instrument = self.arrangement[level]
                        duration = totalDuration * transform.normalizedDuration
                        key = key + transform.deltaKey
                        velocity = velocity + (transform.deltaVelocity * self.gains[level])  
                        phase = 0
                        pan = self.pans[level]
                        print "%2d: %2d %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f" % (level, instrument, time, duration, key, velocity, phase, pan)
                        self.csound.addNote(instrument, time, duration, key, velocity, pan)
                        # Recurse to the next level.
                        self.generate(level + 1, levels, time, duration, key, velocity)
                        time = time + duration
        # Render the generated score.
        def render(self):
                # Ends indefinitely playing effects on the mixer buss.
                self.csound.addScoreLine("e 2")
                # Print the generated score for diagnostic purposes.
                print self.csound.getScore()
                # High-resolution rendering.
                self.csound.setCommand('csound -R -W -Z -f -r 88200 -k 88200 -o tutorial5.py.wav temp.orc temp.sco')
                self.csound.exportForPerformance()
                self.csound.perform()
                self.csound.removeScore()
                
# If running stand-alone, generate a piece;
# if imported by another module, do not generate a piece
# (enables the Generator class to be used as a library).
              
if __name__ == '__main__':

        # Create a generator with four notes 
        # in the same interval relationship as B, A, C, H,
        # i.e. Bb, A, C, B,
        # i.e. 0, -1, +3, -1,
        # offset by a tritone.

        generator = Generator()
        generator.addTransform(10, 6 + 0,  0)
        generator.addTransform( 8,   - 1,  3)
        generator.addTransform( 6,   + 3, -2)
        generator.addTransform(12,   - 1,  0)

        # Generate a 5 minute piece.

        generator.generate(0, 3, 0, 300, 38, 84)
        generator.render()